from .tclab import TCLab, TCLabModel
from .clock import clock, time
from .historian import Historian, Plotter

def setup(connected=True, speedup=1):
    from .clock import setup as clocksetup

    if connected:
        lab = TCLab
        if speedup != 1:
            raise ValueError("The real lab must run in real time")
    else:
        lab = TCLabModel
        if speedup < 0:
            raise ValueError("")

    clocksetup(speedup=speedup)
    return lab


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from __future__ import print_function
from .clock import time
import serial
from serial.tools import list_ports
import random

sep = ' '   # command/value separator in TCLab firmware

arduinos = {'USB VID:PID=16D0:0613': 'Arduino Uno',
            'USB VID:PID=1A86:7523': 'NHduino',
            'USB VID:PID=2341:8036': 'Arduino Leonardo',
            }


def clip(val, lower=0, upper=100):
    """Limit value to be between lower and upper limits"""
    return max(lower, min(val, upper))


class TCLab(object):
    def __init__(self, port='', baud=9600, debug=False):
        self.debug = debug
        print('Connecting to TCLab')
        for comport in list_ports.grep(port):
            if comport[2][:21] in arduinos.keys():
                self.arduino = arduinos[comport[2][:21]]
                break
        else:
            print('--- Serial Ports ---')
            for comport in list(list_ports.comports()):
                print(" ".join(comport))
            raise RuntimeError('No compatible Arduino device found.')
        port = comport[0]
        self.sp = serial.Serial(port=port, baudrate=baud, timeout=2)
        self.receive()
        self.version = self.send_and_receive('VER')
        self._P1 = 200.0
        self._P2 = 100.0
        self.Q1(0)
        self.Q2(0)
        if self.sp.isOpen():
            print(self.version + ' on ' + self.arduino + ' connected to port ' + port)
        self.tstart = time()
        self.sources = [('T1', lambda: self.T1),
                        ('T2', lambda: self.T2),
                        ('Q1', self.Q1),
                        ('Q2', self.Q2),
                        ]

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()
        return

    def close(self):
        """Shut down TCLab device and close serial connection."""
        self.Q1(0)
        self.Q2(0)
        self.send_and_receive('X')
        self.sp.close()
        print('TCLab disconnected successfully.')
        return

    def send(self, msg):
        """Send a string message to the TCLab firmware."""
        self.sp.write((msg + '\r\n').encode())
        if self.debug:
            print('Sent: "' + msg + '"')
        self.sp.flush()

    def receive(self):
        """Return a string message received from the TCLab firmware."""
        msg = self.sp.readline().decode('UTF-8').replace('\r\n', '')
        if self.debug:
            print('Return: "' + msg + '"')
        return msg

    def send_and_receive(self, msg, convert=str):
        """Send a string message and return the response"""
        self.send(msg)
        return convert(self.receive())

    def LED(self, val=100):
        """Flash TCLab LED at a specified brightness (default 100) for 10 seconds."""
        return self.send_and_receive('LED' + sep + str(clip(val)), float)

    @property
    def T1(self):
        """Return a float denoting TCLab temperature T1 in degrees C."""
        return self.send_and_receive('T1', float)

    @property
    def T2(self):
        """Return a float denoting TCLab temperature T2 in degrees C."""
        return self.send_and_receive('T2', float)
    
    @property
    def P1(self):
        """Return a float denoting maximum power of heater 1 in pwm."""
        return self._P1
    
    @P1.setter
    def P1(self, val):
        """Set maximum power of heater 1 in pwm, range 0 to 255."""
        self._P1 = self.send_and_receive('P1' + sep + str(clip(val,0,255)), float)

    @property
    def P2(self):
        """Return a float denoting maximum power of heater 2 in pwm."""
        return self._P2

    @P2.setter
    def P2(self, val):
        """Set maximum power of heater 2 in pwm, range 0 to 255."""
        self._P2 = self.send_and_receive('P2' + sep + str(clip(val,0,255)), float)

    def Q1(self, val=None):
        """Set TCLab heater power Q1 with range limited to 0-100, return clipped value."""
        if val is None:
            msg = 'R1'
        else:
            msg = 'Q1' + sep + str(clip(val))
        return self.send_and_receive(msg, float)

    def Q2(self, val=None):
        """Set TCLab heater power Q1 with range limited to 0-100, return clipped value."""
        if val is None:
            msg = 'R2'
        else:
            msg = 'Q2' + sep + str(clip(val))
        return self.send_and_receive(msg, float)

    # Define properties for Q1 and Q2
    U1 = property(fget=Q1, fset=Q1, doc="Heater 1 value")
    U2 = property(fget=Q2, fset=Q2, doc="Heater 2 value")


class TCLabModel(object):
    def __init__(self, port=None, baud=9600, debug=False):
        self.debug = debug
        print('Simulated TCLab')
        self.Ta = 21                  # ambient temperature
        self.tstart = time()          # start time
        self.tlast = self.tstart      # last update time
        self._P1 = 200.0              # max power heater 1
        self._P2 = 100.0              # max power heater 2
        self._Q1 = 0                  # initial heater 1
        self._Q2 = 0                  # initial heater 2
        self._T1 = self.Ta            # temperature thermister 1
        self._T2 = self.Ta            # temperature thermister 2
        self._H1 = self.Ta            # temperature heater 1
        self._H2 = self.Ta            # temperature heater 2
        self.maxstep = 0.2            # maximum time step for integration
        self.sources = [('T1', lambda: self.T1),
                        ('T2', lambda: self.T2),
                        ('Q1', self.Q1),
                        ('Q2', self.Q2),
                        ]

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()
        return

    def close(self):
        """Simulate shutting down TCLab device."""
        self.Q1(0)
        self.Q2(0)
        print('Surrogate TCLab disconnected successfully.')
        return

    def LED(self, val=100):
        """Simulate flashing TCLab LED

           val : specified brightness (default 100). """
        self.update()
        return clip(val)

    @property
    def T1(self):
        """Return a float denoting TCLab temperature T1 in degrees C."""
        self.update()
        return self._T1 + random.normalvariate(0, 0.2)

    @property
    def T2(self):
        """Return a float denoting TCLab temperature T2 in degrees C."""
        self.update()
        return self._T2 + random.normalvariate(0, 0.2)

    @property
    def P1(self):
        """Return a float denoting maximum power of heater 1 in pwm."""
        self.update()
        return self._P1
    
    @P1.setter
    def P1(self, val):
        """Set maximum power of heater 1 in pwm, range 0 to 255."""
        self.update()
        self._P1 = clip(val,0,255)

    @property
    def P2(self):
        """Return a float denoting maximum power of heater 2 in pwm."""
        self.update()
        return self._P2

    @P2.setter
    def P2(self, val):
        """Set maximum power of heater 2 in pwm, range 0 to 255."""
        self.update()
        self._P2 = clip(val,0,255)

    def Q1(self, val=None):
        """Simulate setting TCLab heater power Q1 with range limited to 0-100, return clipped value."""
        self.update()
        if val is not None:
            self._Q1 = clip(val)
        return self._Q1

    def Q2(self, val=None):
        """Simulate setting TCLab heater power Q2 with range limited to 0-100, return clipped value."""
        self.update()
        if val is not None:
            self._Q2 = clip(val)
        return self._Q2

    # Define properties for Q1 and Q2
    U1 = property(fget=Q1, fset=Q1, doc="Heater 1 value")
    U2 = property(fget=Q2, fset=Q2, doc="Heater 2 value")

    def update(self):
        # Time updates
        self.tnow = time() - self.tstart
        trequired = self.tnow - self.tlast
        self.tlast = self.tnow

        fullsteps, remainder = divmod(trequired, self.maxstep)
        steps = [self.maxstep]*int(fullsteps) + [remainder]

        for dt in steps:
            DeltaTaH1 = self.Ta - self._H1
            DeltaTaH2 = self.Ta - self._H2
            DeltaT12 = self._H1 - self._H2
            dH1 = self._P1 * self._Q1 / 5720 + DeltaTaH1 / 20 - DeltaT12 / 100
            dH2 = self._P2 * self._Q2 / 5720 + DeltaTaH2 / 20 + DeltaT12 / 100
            dT1 = (self._H1 - self._T1)/140
            dT2 = (self._H2 - self._T2)/140

            self._H1 += dt * dH1
            self._H2 += dt * dH2
            self._T1 += dt * dT1
            self._T2 += dt * dT2



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from __future__ import print_function
from __future__ import division
from .clock import time
import bisect

class Historian(object):
    """Generalised logging class"""
    def __init__(self, sources):
        """

        sources: an iterable of (name, callable) tuples
                     - name (str) is the name of a signal and the
                     - callable is evaluated to obtain the value
        """
        self.sources = [('Time', lambda: self.tnow)] + list(sources)
        self.tstart = time()
        self.tnow = 0
        self.columns = [name for name, _ in self.sources]
        self.fields = [[] for _ in range(len(self.sources))]
        self.logdict = {c: f for c, f in zip(self.columns, self.fields)}
        self.t = self.logdict['Time']

        self.update(tnow=0)

    def update(self, tnow=None):
        if tnow is None:
            self.tnow = time()
        else:
            self.tnow = tnow

        for n, c in self.sources:
            self.logdict[n].append(c())

    @property
    def log(self):
        return list(zip(*[self.logdict[c] for c in self.columns]))

    def at(self, t, columns=None):
        """ Return the values of columns at or just before a certain time"""
        if columns is None:
            columns = self.columns
        i = bisect.bisect(self.t, t) - 1
        return [self.logdict[c][i] for c in columns]


class Plotter:
    def __init__(self, historian, tperiod=20):
        self.historian = historian

        import matplotlib.pyplot as plt
        from IPython import display

        display.clear_output()
        self.plt = plt
        self.display = display

        line_options = {'lw': 2, 'alpha': 0.8}

        self.fig = plt.figure(figsize=(8, 6))
        nplots = len(self.historian.columns) - 1  
        self.lines = []
        self.axes = []
        for n in range(0, nplots):
            self.axes.append(self.fig.add_subplot(nplots,1,n+1))
            y = self.historian.at(0,[self.historian.columns[n+1]])[0]
            li, = plt.step(0, y, where='post', **line_options)          
            self.lines.append(li)
            plt.xlim(0, 1.05 * tperiod)
            plt.ylim(y-2,y+2)
            plt.ylabel(self.historian.columns[n+1])
            plt.grid()
        plt.xlabel('Time / Seconds')
        plt.tight_layout()
        display.display(plt.gcf())

    def update(self, tnow=None):
        self.historian.update(tnow)

        plt = self.plt
        display = self.display

        t = self.historian.fields[0]
        for n in range(0, len(self.historian.columns)-1):
            y = self.historian.fields[n+1]
            self.lines[n].set_data(t,y)
            ymax = max(y)
            ymin = min(y)
            if (ymax > self.axes[n].get_ylim()[1]) or (ymin < self.axes[n].get_ylim()[0]):
                self.axes[n].set_ylim(ymin-1,ymax+2)
            if self.historian.tnow > self.axes[n].get_xlim()[1]:
                self.axes[n].set_xlim(0, 1.4 * self.axes[n].get_xlim()[1])
        display.clear_output(wait=True)
        display.display(plt.gcf())




import time as original_time
import gc

SPEEDUP = 1
tnow = 0


def setup(speedup=1):
    global SPEEDUP
    SPEEDUP = speedup


def time():
    """Returns current clock time."""
    return tnow


def speedtime():
    """Returns time rescaled by factor SPEEDUP."""
    return SPEEDUP*original_time.time()


def speedsleep(tsleep):
    """Sleeps for a period tsleep rescaled by factor SPEEDUP."""
    return original_time.sleep(tsleep/SPEEDUP)


def clock(tperiod, tstep=1, strict=True, tol=0.1):
    """Generator providing time values in sync with real time clock.

    Args:
        tperiod (float): Time interval for clock operation in seconds.
        tstep (float): Time step.
        strict (bool): False turns off real-time synchronization.
        tol (float): Maximum permissible deviation from real time.

    Yields:
        float: The next time step rounded to nearest 10th of a second.

    Raises:
         TCLabClockError: If clock becomes more than `tol` out of phase
             with real time clock.
    """
    global tnow
    tnow = 0
    start_time = speedtime()
    fuzz = 0.003
    k = 0
    strict = False if SPEEDUP > 1 else strict
    while tnow <= tperiod - tstep + tol + fuzz:
        yield round(tnow, 1)
        if SPEEDUP < 10:
            if strict:
                tsleep = max(0, (k+1)*tstep - (speedtime() - start_time) - fuzz*SPEEDUP)
            else:
                tsleep = max(0, tstep - (speedtime() - start_time - tnow) - fuzz*SPEEDUP)
            gcold = gc.isenabled()
            gc.disable()
            try:
                if tsleep >= fuzz:
                    speedsleep(tsleep)
            finally:
                if gcold:
                    gc.enable()
            tnow = speedtime() - start_time
        else:
            tnow = (k+1)*tstep
        k += 1
        if strict and (abs(tnow - k * tstep) > tol + fuzz):
            raise RuntimeError("TCLab clock lost real time synchronization.")

    yield round(tnow, 1)
